/*
** Module   : EditorWidget.java
** Abstract : server side Editor widget implementation
**
** Copyright (c) 2006-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 SIY 20060427   @26084 Server side part of the EDITOR widget.
** 002 GES 20060803   @28373 Cleanup and refactoring.
** 003 GES 20080320   @37585 Added TYPE attribute.
** 004 SIY 20111101          Updated imports.
** 005 CA  20130129          Added setLarge stub, for EDITOR LARGE.
** 006 CA  20130221          Refactored usage of TYPE attribute.
** 007 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 008 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 009 CA  20131115          Changes related to properly instantiate dynamic resources.
** 010 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 011 SVL 20141102          Make this class implement ScrollbarVerticalElement.
** 012 GES 20141124          Added EditorInterface support (with all associated attrs/meths).
** 013 GES 20150103          Added runtime support for new attrs/methods.
** 014 EVK 20150209          Added runtime support for save-file and read-file methods.
** 015 IAS 20150318          Added runtime support for new attrs/methods (
**                           BOX, RETURN-INSERTED. CURSOR-XXX, REPLACE...)
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Server side implementation of the EDITOR widget.
 */
@LegacyResource(resource = LegacyResource.EDITOR)
public class EditorWidget
extends ControlTextWidget<EditorConfig>
implements EditorInterface,
           BoxInterface,
           ScrollbarVerticalElement,
           ScrollbarHorizontalElement,
           InnerLines
{
   /**
    * Default constructor.
    */
   public EditorWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public EditorWidget(boolean dynamic)
   {
      super(dynamic, new EditorConfig());
   }

   /**
    * Mark this EDITOR widget as EDITOR LARGE, which can be used to display a value of type 
    * <code>LONGCHAR</code>.
    * 
    * @param    large
    *           When <code>true</code>, this is a EDITOR LARGE widget.
    */
   public void setLarge(boolean large)
   {
      UnimplementedFeature.missing("setLarge() method");
   }
   
   /**
    * Gets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @return   The number of lines of text.
    */
   public integer getInnerLines()
   {
      return new integer(config.innerLines);
   }
   
   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   public void setInnerLines(double lines)
   {
      config.innerLines = (int) lines;
      pushScreenDefinition();
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   public void setInnerLines(NumberType lines)
   {
      if (lines == null || lines.isUnknown())
      {
         UnimplementedFeature.todo("INNER-LINES attribute unknown behavior is not known.");
         return;
      }
   
      setInnerLines(lines.intValue());
   }

   /**
    * Gets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @return   The number of character columns.
    */
   public integer getInnerChars()
   {
      return new integer(config.innerChars);
   }
   
   /**
    * Sets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setInnerChars(double chars)
   {
      config.innerChars = (int) chars;
      pushScreenDefinition();
   }
   
   /**
    * Sets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setInnerChars(NumberType chars)
   {
      if (chars == null || chars.isUnknown())
         return;
      
      setInnerChars(chars.intValue());
   }

   /**
    * Set value of the MAX-CHARS attribute.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setMaxChars(int chars)
   {
      config.maxChars = chars;
      pushScreenDefinition();
   }

   /**
    * Set value of the MAX-CHARS attribute.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setMaxChars(NumberType chars)
   {
      if (chars.isUnknown())
         return;
      
      setMaxChars(chars.intValue());
   }
   
   /**
    * Set value of the BUFFER-CHARS attribute.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setBufferChars(int chars)
   {
      config.bufferChars = chars;
      pushScreenDefinition();
   }
   
   /**
    * Set value of the BUFFER-CHARS attribute.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setBufferChars(NumberType chars)
   {
      if (chars.isUnknown())
         return;
      
      setBufferChars(chars.intValue());
   }
   
   /**
    * Set value of the BUFFER-LINES attribute.
    * 
    * @param   lines
    *          The new value for the attribute.
    */
   public void setBufferLines(int lines)
   {
      config.bufferLines = lines;
      pushScreenDefinition();
   }
   
   /**
    * Set value of the BUFFER-LINES attribute.
    * 
    * @param   lines
    *          The new value for the attribute.
    */
   public void setBufferLines(NumberType lines)
   {
      if (lines.isUnknown())
         return;
      
      setBufferLines(lines.intValue());
   }
   
   /**
    * Set value of the BOX attribute as opposite to passed value.
    * 
    * @param   box
    *          The new value for the attribute.
    */
   public void setNoBox(boolean box)
   {
      config.box = !box;
      pushScreenDefinition();
   }

   /**
    * Set value of the BOX attribute as opposite to passed value.
    * 
    * @param   box
    *          The new value for the attribute.
    */
   public void setNoBox(logical box)
   {
      if (box.isUnknown())
         return;
      
      setNoBox(box.booleanValue());
   }
   
   /**
    * Set value of the WORD-WRAP attribute as opposite to passed value.
    * 
    * @param   wordwrap
    *          The new value for the attribute.
    */
   public void setNoWordWrap(boolean wordwrap)
   {
      config.wordWrap = !wordwrap;
      pushScreenDefinition();
   }

   /**
    * Set value of the WORD-WRAP attribute as opposite to passed value.
    * 
    * @param   wordwrap
    *          The new value for the attribute.
    */
   public void setNoWordWrap(logical wordwrap)
   {
      if (wordwrap.isUnknown())
         return;
      
      setNoWordWrap(wordwrap.booleanValue());
   }
   
   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param   vertical
    *          The new value for the attribute.
    */
   public void setScrollbarVertical(boolean vertical)
   {
      config.scrollbarVertical = vertical;
      pushScreenDefinition();
   }

   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param   vertical
    *          The new value for the attribute.
    */
   public void setScrollbarVertical(logical vertical)
   {
      if (vertical == null || vertical.isUnknown())
      {
         UnimplementedFeature.todo("SCROLLBAR-VERTICAL attribute unknown behavior is not " +
                                   "known.");
         return;
      }
      
      setScrollbarVertical(vertical.booleanValue());
   }

   /**
    * Get value of the SCROLLBAR-VERTICAL attribute.
    *
    * @return  value of the attribute.
    */
   public logical isScrollbarVertical()
   {
      return new logical(config.scrollbarVertical);
   }
   
   /**
    * Get the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    * 
    * @return   The value of the SCROLLBAR-HORIZONTAL attribute.
    */
   public logical isScrollbarHorizontal()
   {
      return new logical(config.scrollbarHorizontal);
   }
   
   /**
    * Set the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    *
    * @param   horiz
    *          The new value for the attribute.
    */
   public void setScrollbarHorizontal(logical horiz)
   {
      if (horiz == null || horiz.isUnknown())
      {
         UnimplementedFeature.todo("SCROLLBAR-HORIZONTAL attribute unknown behavior is not " +
                                   "known.");
         return;
      }
      
      setScrollbarHorizontal(horiz.booleanValue());
   }
   
   /**
    * Set the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    *
    * @param   horiz
    *          The new value for the attribute.
    */
   public void setScrollbarHorizontal(boolean horiz)
   {
      config.scrollbarHorizontal = horiz;
      pushScreenDefinition();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int64 end)
   {
      if (start == null || start.isUnknown())
      {
         config.selectionActive = false;
         return new logical(true);
      }
      return setSelection(start.intValue(), end);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int end)
   {
      if (start == null || start.isUnknown())
      {
         config.selectionActive = false;
         return new logical(true);
      }
      return setSelection(start.intValue(), end);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int64 end)
   {
      if (end == null || end.isUnknown())
      {
         config.selectionActive = false;
         return new logical(true);
      }
      return setSelection(start, end.intValue());
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int end)
   {
      return new logical(LogicalTerminal.setEditorSelection(getId(), start, end));
   }
   
   /**
    * Implements the DELETE-LINE() method which deletes the line of text in which the cursor is
    * positioned.
    *
    * @return   <code>true</code> if the line was deleted.
    */
   public logical deleteLine()
   {
      return new logical(LogicalTerminal.deleteEditorLine(getId()));
   }
   
   /**
    * Implements the INSERT-STRING() method which inserts the given text into the editor at the
    * location at which the cursor is positioned.
    *
    * @param    txt
    *           The text to be inserted.
    *
    * @return   <code>true</code> if the text was inserted.
    */
   public logical insertString(character txt)
   {
      return insertString(txt == null || txt.isUnknown() ? "" : txt.toStringMessage());
   }
   
   /**
    * Implements the INSERT-STRING() method which inserts the given text into the editor at the
    * location at which the cursor is positioned.
    *
    * @param    txt
    *           The text to be inserted.
    *
    * @return   <code>true</code> if the text was inserted.
    */
   public logical insertString(String txt)
   {
      return new logical(
                  LogicalTerminal.insertEditorString(getId(), txt == null ? "" : txt)
               );
   }
   
   /**
    * Implements the READ-FILE() method which replaces the editor's contents with those read
    * from the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully read.
    */
   public logical readFile(character name)
   {
      if (name == null || name.isUnknown())
      {
         return new logical(false);
      }
      
      boolean res = LogicalTerminal.editorReadFile(config.id.getId(), name.toStringMessage());
      
      return new logical(res);
   }
   
   /**
    * Implements the READ-FILE() method which replaces the editor's contents with those read
    * from the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully read.
    */
   public logical readFile(String name)
   {
      return readFile(new character(name));
   }
   
   /**
    * Implements the REPLACE-SELECTION-TEXT() method which replaces the editor's currently
    * selected text with the given text.
    *
    * @param    txt
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   public logical replaceSelectionText(character txt)
   {
      return replaceSelectionText(txt == null || txt.isUnknown() ? "" : txt.toStringMessage());
   }
   
   /**
    * Implements the REPLACE-SELECTION-TEXT() method which replaces the editor's currently
    * selected text with the given text.
    *
    * @param    txt
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   public logical replaceSelectionText(String txt)
   {
      if (!config.selectionActive)
      {
         return new logical(false);
      }
      return new logical(
               LogicalTerminal.replaceEditorSelectionText(getId(), txt == null ? "" : txt)
            );
   }

   /**
    * Implements the SAVE-FILE() method which write the editor's contents to the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully written.
    */
   @Override
   public logical saveFile(String name)
   {
      return saveFile(new character(name));
   }

   /**
    * Implements the SAVE-FILE() method which write the editor's contents to the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully written.
    */
   @Override
   public logical saveFile(character name)
   {
      if (name == null || name.isUnknown())
      {
         return new logical(false);
      }

      boolean res = LogicalTerminal.editorSaveFile(config.id.getId(), name.toStringMessage());

      return new logical(res);
   }

   /**
    * Get the value of the CURSOR-CHAR attribute which is the current cursor position as a
    * 1-based index on the current line.
    *
    * @return   The current position of the cursor on the current line.
    */
   public integer getCursorChar()
   {
      return new integer(config.cursorChar);
   }
   
   /**
    * Set the value of the CURSOR-CHAR attribute which moves the cursor position to the
    * specified 1-based index on the current line.
    *
    * @param    idx
    *           The 1-based position to which to move.
    */
   public void setCursorChar(double idx)
   {
      if (idx < 1)
      {
         invalidPosition("CURSOR-CHAR", 0);
         return;
      }

      Boolean result = LogicalTerminal.setEditorCursorChar(getId(), (int)idx);
      if (result == null)
      {
         UnimplementedFeature.todo("CURSOR-CHAR return false compatibility testing is needed");
      }
      else if (!result.booleanValue())
      {
         invalidPosition("CURSOR-CHAR", (int)idx);
      }
   }

   /**
    * Set the value of the CURSOR-CHAR attribute which moves the cursor position to the
    * specified 1-based index on the current line.
    *
    * @param    idx
    *           The 1-based position to which to move.
    */
   public void setCursorChar(NumberType idx)
   {
      if (idx == null || idx.isUnknown())
      {
         invalidPosition("CURSOR-CHAR", 0);
         return;
      }
      setCursorChar(idx.doubleValue()); 
   }

   /**
    * Get the value of the CURSOR-LINE attribute which is the current cursor position as a
    * 1-based line number.
    *
    * @return   The current position of the cursor on the current line.
    */
   public integer getCursorLine()
   {
      return new integer(config.cursorY + config.scrollY + 1);
   }
   
   /**
    * Set the value of the CURSOR-LINE attribute which moves the cursor position to the
    * specified 1-based line number.
    *
    * @param    line
    *           The line to which to move.
    */
   public void setCursorLine(double line)
   {
      if (line < 1)
      {
         invalidPosition("CURSOR-LINE", 0);
         return;
      }
      
      Boolean result = LogicalTerminal.setEditorCursorLine(getId(), (int)line);
      if (result == null)
      {
         UnimplementedFeature.todo("CURSOR-LINE return false compatibility testing is needed");
      }
      else if (!result.booleanValue())
      {
         invalidPosition("CURSOR-LINE", (int)line);
      }
   }

   /**
    * Set the value of the CURSOR-LINE attribute which moves the cursor position to the
    * specified 1-based line number.
    *
    * @param    line
    *           The line to which to move.
    */
   public void setCursorLine(NumberType line)
   {
      if (line == null || line.isUnknown())
      {
         invalidPosition("CURSOR-LINE", 0);
         return;
      }
      
      setCursorLine(line.doubleValue()); 
   }
   
   /**
    * Obtains the index position of the cursor in a text editing widget. This corresponds to
    * the CURSOR-OFFSET attribute in the  EDITOR widget.
    * 
    * @return   The character index at which the cursor is placed.
    */
   @Override
   public integer getCursorOffset()
   {
      return new integer(config.cursorOffset);
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute in the EDITOR widget.
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(double offset)
   {
      if (offset < 1)
      {
         invalidPosition("CURSOR-OFFSET", 0);
         return;
      }

      Boolean result = LogicalTerminal.setEditorCursorOffset(getId(), (int)offset);
      if (result == null)
      {
         UnimplementedFeature.todo("CURSOR-OFFSET return false compatibility testing is needed");
      }
      else if (!result.booleanValue())
      {
         invalidPosition("CURSOR-OFFSET", (int)offset);
      }
 }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute in the EDITOR widget.
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(NumberType offset)
   {
      if (offset == null || offset.isUnknown())
      {
         invalidPosition("CURSOR-OFFSET", 0);
         return;
      }
      
      setCursorOffset(offset.doubleValue()); 
   }

   /**
    * Get the value of the NUM-LINES attribute which is the number of lines of text in the
    * editor.
    *
    * @return   The number of lines in the editor.
    */
   public integer getNumLines()
   {
      return new integer(config.numLines);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, character to, integer flag)
   {
      if (from == null || from.isUnknown())
      {
         // 4GL hangs in this case 
         return new logical(false);
      }
      return replace(from.toStringMessage(), to, flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(String from, character to, integer flag)
   {
      return replace(from, to == null || to.isUnknown() ? "" : to.toStringMessage(), flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, String to, integer flag)
   {
      if (from == null || from.isUnknown())
      {
         // 4GL hangs in this case 
         return new logical(false);
      }
      return replace(from.toStringMessage(), to, flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, character to, int flag)
   {
      if (from == null || from.isUnknown())
      {
         // 4GL hangs in this case 
         return new logical(false);
      }
      return replace(from.toStringMessage(), to, flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(String from, character to, int flag)
   {
      return replace(from, to == null || to.isUnknown() ? "" : to.toStringMessage(), flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, String to, int flag)
   {
      if (from == null || from.isUnknown())
      {
         // 4GL hangs in this case 
         return new logical(false);
      }
      return replace(from.toStringMessage(), to, flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(String from, String to, integer flag)
   {
      if (flag == null || flag.isUnknown())
      {
         return new logical(false);
      }
      return replace(from, to, flag.intValue());
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(String from, String to, int flag)
   {
      if (from == null || "".equals(from))
      {
         // 4GL crashes in this case 
         return new logical(false);
      }

      Boolean result = LogicalTerminal.replaceEditorString(getId(), from, to, flag);
      if (result == null)
         UnimplementedFeature.todo("REPLACE() return false compatibility testing is needed");
      return new logical(result);
   }

   /**
    * Get the value of the RETURN-INSERTED attribute
    *
    * @return   the value of the attribute
    */
   @Override
   public logical isReturnInserted()
   {
      return new logical(config.returnInserted);
   }

   /**
    * Set the value of the RETURN-INSERTED attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setReturnInserted(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute RETURN-INSERTED on %s %s", 
                  type(), frame.getName(this)
               ), 
               false);
      }
      setReturnInserted(value.booleanValue());
   }

   /**
    * Set the value of the RETURN-INSERTED attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setReturnInserted(boolean value)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set RETURN-INSERTED because the %s %s has been realised", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }
      config.returnInserted = value;
   }

   /**
    * Get the value of the BOX attribute
    *
    * @return   the value of the attribute
    */
   @Override
   public logical isBox()
   {
      return new logical(config.box);
   }

   /**
    * Set the value of the BOX attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setBox(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute BOX on %s %s", 
                  type(), frame.getName(this)
               ), 
               false);
         
      }
      setBox(value.booleanValue());
   }

   /**
    * Set the value of the BOX attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setBox(boolean value)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set BOX because the %s %s has been realised", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }
      config.box = value;
      pushScreenDefinition();
  }

   /**
    * Report invalid cursor positioning attribute 
    *
    * @param attr
    *          attribute name
    * @param value
    *          attribute value
    */
   private void invalidPosition(String attr, int value)
   {
      ErrorManager.recordOrShowError(4057, 
            String.format(
               "Attribute %s for the %s %s has an invalid value of %d", 
               attr, type(), frame.getName(this), value
            ), 
            false);
   }

   /**
    * Get the value of the LENGTH attribute
    *
    * @return   the value of the attribute
    */
   @Override
   public integer getLength()
   {
      return new integer(config.length);
   }

   /**
    * Get SELECTION-TEXT attribute.
    *
    * @return  SELECTION-TEXT value   
    */
   @Override
   public character getSelectionText()
   {
      if (!config.selectionActive)
      {
         return new character("");
      }
      return new character(config.selectionText);
   }

   /**
    * Get SELECTION-END attribute.
    *
    * @return  SELECTION-END value   
    */
   @Override
   public integer getSelectionEnd()
   {
      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionEnd);
   }

   /**
    * Get SELECTION-START attribute.
    *
    * @return  SELECTION-START value   
    */
   @Override
   public integer getSelectionStart()
   {
      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionStart);
   }

   /**
    * Get TEXT-SELECTED attribute.
    *
    * @return  TEXT-SELECTED value   
    */
   @Override
   public logical isTextSelected()
   {
      return new logical(config.textSelected);
   }
   
   
}
